@@ -10,7 +10,10 @@ use Selenium::Remote::WDKeys ();
 
 use constant DEBUG => $ENV{MOJO_SELENIUM_DEBUG} || 0;
 
-our $VERSION = '0.05';
+$ENV{TEST_SELENIUM} //= '0';
+$ENV{MOJO_SELENIUM_BASE_URL} ||= $ENV{TEST_SELENIUM} =~ /^http/ ? $ENV{TEST_SELENIUM} : '';
+
+our $VERSION = '0.06';
 
 my $SCRIPT_NAME = File::Basename::basename($0);
 my $SCREENSHOT  = 1;
@@ -198,31 +201,31 @@ sub navigate_ok {
   return $self->_test('ok', !$err, _desc($desc));
 }
 
-around new => sub {
-  my $next = shift;
-  my $self = $next->(@_);
-  my $app  = $self->app;
+sub new {
+  my $self = shift->SUPER::new;
   $self->ua(Test::Mojo::Role::Selenium::UserAgent->new->ioloop(Mojo::IOLoop->singleton));
-  $self->app($app) if $app;
-  return $self;
-};
+  return $self if $ENV{MOJO_SELENIUM_BASE_URL};
+  return $self unless my $app = shift;
+  return $self->app(ref $app ? $app : Mojo::Server->new->build_app($app));
+}
 
 sub refresh { $_[0]->_proxy('refresh'); $_[0] }
 
 sub send_keys_ok {
   my ($self, $selector, $keys, $desc) = @_;
-  my $el = $self->_proxy(find_element => $selector);
-  my @keys;
+  my $el = $selector ? $self->_proxy(find_element => $selector) : $self->driver->get_active_element;
 
-  for (ref $keys ? @$keys : ($keys)) {
+  $selector ||= 'active element';
+
+  for (@$keys) {
     my $key = ref $_ ? Selenium::Remote::WDKeys::KEYS()->{$$_} : $_;
     croak "Invalid key '@{[ref $_ ? $$_ : $_]}'" unless defined $key;
-    push @keys, $key;
+    $_ = $key;
   }
 
   if ($el) {
     eval {
-      for my $key (@keys) {
+      for my $key (@$keys) {
         warn "[Selenium] send_keys $selector <- @{[Mojo::Util::url_escape($key)]}\n" if DEBUG;
         $el->send_keys($key);
       }
@@ -249,7 +252,7 @@ sub setup_or_skip_all {
   Test::More::plan(skip_all => $@ || 'TEST_SELENIUM=1 or TEST_SELENIUM=http://...')
     unless $ENV{TEST_SELENIUM} and eval { $self->driver };
 
-  $ENV{MOJO_SELENIUM_BASE_URL} //= $ENV{TEST_SELENIUM} if $ENV{TEST_SELENIUM} =~ /^http/;
+  $ENV{MOJO_SELENIUM_BASE_URL} ||= $ENV{TEST_SELENIUM} if $ENV{TEST_SELENIUM} =~ /^http/;
 
   return $self;
 }
@@ -261,21 +264,32 @@ sub submit_ok {
   return $self->_test('ok', $el, _desc($desc, "click on $selector"));
 }
 
-sub wait_for_url {
-  my ($self, $regex, $desc) = @_;
-  my $driver = $self->driver;
-  my $guard  = 600;             # 60 seconds
-  my $ok;
+sub wait_until {
+  my ($self, $cb, $args) = @_;
+  my $ioloop = $self->ua->ioloop;
+  my @tid;
 
-  while ($guard--) {
-    Time::HiRes::usleep(1e5);
-    warn "[Selenium] current_url=@{[$driver->get_current_url]}\n" if DEBUG > 2;
-    next unless $driver->get_current_url =~ /$regex/;
-    $ok = 1;
-    last;
-  }
+  $args->{timeout}  ||= $ENV{MOJO_SELENIUM_WAIT_TIMEOUT}  || 60;
+  $args->{interval} ||= $ENV{MOJO_SELENIUM_WAIT_INTERVAL} || 0.5;
 
-  return $self->_test('ok', $ok, _desc($desc, 'got expected url'));
+  $ioloop->delay(
+    sub {
+      my $next = shift->begin;
+      push @tid, $ioloop->timer($args->{timeout}, $next);
+      push @tid, $ioloop->recurring(
+        $args->{interval},
+        sub {
+          $next->() if eval { local $_ = $self->driver; $self->$cb($args) };
+          Test::More::diag("[Selenium] wait_until: $@") if $@ and ($args->{debug} or DEBUG);
+        }
+      );
+    },
+    sub {
+      $ioloop->remove($_) for @tid;
+    },
+  )->wait;
+
+  return $self;
 }
 
 sub window_size_is {
@@ -329,12 +343,29 @@ use constant DEBUG => $ENV{MOJO_SELENIUM_DEBUG} || 0;
 
 sub request {
   my ($ua, $req) = @_;
-  my $method = lc $req->method || 'get';
-  warn "[Selenium] @{[uc $method]} @{[$req->uri->as_string]}\n" if DEBUG;
-  my $tx = $ua->$method($req->uri->as_string, {$req->headers->flatten}, $req->content);
+  my $method = uc($req->method || 'get');
+  my $tx = $ua->build_tx($method, $req->uri->as_string, {$req->headers->flatten}, $req->content);
+  my $done;
+
+  warn "[Selenium] $method @{[$req->uri->as_string]}\n" if DEBUG;
+
+  # This is super ugly and need to be implemented differently,
+  # but I'm not sure how to implement wait_until() without this
+  # one_tick() hack.
+  if ($ua->ioloop->is_running) {
+    $ua->start($tx, sub { $done = 1 });
+    $ua->ioloop->reactor->one_tick until $done;
+  }
+  else {
+    $ua->start($tx);
+  }
+
   return HTTP::Response->parse($tx->res->to_string);
 }
 
+# Should not say "... during global destruction."
+# sub DESTROY { warn 'no circular refs?' }
+
 1;
 
 =encoding utf8
@@ -362,7 +393,7 @@ Test::Mojo::Role::Selenium - Test::Mojo in a real browser
   $t->element_is_displayed("input[name=q]")
     ->send_keys_ok("input[name=q]", ["render", \"return"]);
 
-  $t->wait_for_url(qr{q=render})
+  $t->wait_until(sub { $_->get_current_url =~ qr{q=render} })
     ->live_value_is("input[name=search]", "render");
 
   done_testing;
@@ -517,8 +548,10 @@ input format. The format supports these special strings:
 =head2 click_ok
 
   $self = $self->click_ok("a");
+  $self = $self->click_ok;
 
-Click on an element.
+Click on an element matching the selector or click on the currently active
+element.
 
 =head2 current_url_is
 
@@ -639,6 +672,14 @@ matches the given regex.
 
 Open a browser window and go to the given location.
 
+=head2 new
+
+  $self = $class->new;
+  $self = $class->new($app);
+
+Same as L<Test::Mojo/new>, but will not build C<$app> if
+L</MOJO_SELENIUM_BASE_URL> is set.
+
 =head2 refresh
 
   $self = $self->refresh;
@@ -649,11 +690,12 @@ See L<Selenium::Remote::Driver/refresh>.
 
 =head2 send_keys_ok
 
-  $self->send_keys_ok("input[name=username]", "jhthorsen");
-  $self->send_keys_ok("input[name=name]", ["jan", \"space", "henning"]);
+  $self->send_keys_ok("input[name=name]", ["web", \"space", "framework"]);
+  $self->send_keys_ok(undef, [\"return"]);
 
 Used to send keys to a given element. Scalar refs will be sent as
-L<Selenium::Remote::WDKeys> strings.
+L<Selenium::Remote::WDKeys> strings. Passing in C<undef> as the first argument
+will cause the keys to be sent to the currently active element.
 
 List of some of the special keys:
 
@@ -695,12 +737,13 @@ Submit a form, either by selector or the current active form.
 
 See L<Selenium::Remote::WebElement/submit>.
 
-=head2 wait_for_url
+=head2 wait_until
 
-  $self = $self->wait_for_url(qr{/whatever});
+  $self = $self->wait_until(sub { my $self = shift; return 1 });
+  $self = $self->wait_until(sub { $_->get_current_url =~ /foo/ });
 
-Will sleep until the URL in the browser changes to something that matches the
-given regex.
+Start L<Mojo::IOLoop> and run it until the callback returns true. Note that
+C<$_[0]> is C<$self> and C<$_> is L</driver>.
 
 =head2 window_size_is
 
